use gloo::console::__macro::JsValue;
use gloo::file::callbacks::read_as_text;
use gloo::file::callbacks::FileReader;
use gloo::file::File;

use serde::Deserialize;

use web_sys::Event;
use web_sys::FileList;
use web_sys::HtmlInputElement;

use yew::html;
use yew::html::TargetCast;
use yew::Component;
use yew::Context;
use yew::Html;

const CHALLENGES: &[&str] = &[
    "Beginner's Gauntlet",
    "Battlebot Battle Royale",
    "Hero Battle Royale",
    "Kalaw's Challenge",
    "Army of One",
    "Ring-Out Challenge",
    "Flip Fire Gauntlet",
    "Vanishing Maze",
    "Mondo Condo",
    "Birds of Prey",
    "Battlebot Revenge",
    "Mach Speed Melee",
    "Galactic Rumble",
    "Stop and Go",
    "Mecha Madness",
    "Rolling Thunder",
    "Blast from the Past",
    "Bubble Battle",
];

#[derive(Deserialize, Debug)]
struct FPSave {
    #[serde(rename = "challengeRank")]
    challenge_rank: Vec<i8>,
}

struct FileDetails {
    name: String,
    data: String,
}

pub enum Msg {
    Failure(String),
    Loaded(String, String),
    Upload(File),
}

#[derive(Default)]
pub struct App {
    reader: Option<FileReader>,
    file: Option<FileDetails>,
}

impl Component for App {
    type Message = Msg;
    type Properties = ();

    fn create(_ctx: &Context<Self>) -> Self {
        Self::default()
    }

    fn update(&mut self, ctx: &Context<Self>, msg: Self::Message) -> bool {
        match msg {
            Msg::Loaded(name, data) => self.loaded(name, data),
            Msg::Upload(file) => self.upload(file, ctx),
            Msg::Failure(msg) => {
                println!("{msg}");
                false
            }
        }
    }

    fn view(&self, ctx: &Context<Self>) -> Html {
        html! {
            <div id="wrapper">
                <p id="title">
                    { "Freedom Planet Challenges Rank Viewer" }
                </p>
                <div id="preview-area">
                    { for self.file.iter().map(Self::view_file) }
                </div>
                <label for="file-upload">
                    <p>{">> Click here to select a save file <<"}</p>
                </label>
                <p>
                    {"Based on the "}
                    <a href="https://github.com/yewstack/yew/tree/master/examples/file_upload">
                        {"Yew File Upload"}
                    </a>
                    {" example"}
                </p>
                <input
                    id="file-upload"
                    type="file"
                    accept="application/json"
                    onchange={ctx.link().callback(move |e: Event| {
                        let input: HtmlInputElement = e.target_unchecked_into();
                        Self::process(input.files())
                    })}
                />
            </div>
        }
    }
}

impl App {
    fn view_file(file: &FileDetails) -> Html {
        let Ok(save): Result<FPSave, _> = serde_json::from_str(&file.data) else {
            return html! { <p>{"Invalid save file!"}</p> };
        };

        let ranks = save
            .challenge_rank
            .iter()
            .copied()
            .skip(1)
            .take(CHALLENGES.len())
            .map(Self::rank_text)
            .collect::<Vec<_>>();

        let challenges = CHALLENGES
            .iter()
            .copied()
            .zip(ranks.into_iter())
            .map(|(chal, rank)| Self::view_challenge(chal, rank));

        html! {
            <div class="preview-media">
                <table>
                    <thead>
                        <tr>
                            <th>{ "Challenge" }</th>
                            <th>{ "Rank" }</th>
                        </tr>
                    </thead>
                    <tbody>
                        { for challenges }
                    </tbody>
                </table>
            </div>
        }
    }

    fn view_challenge(challenge: &str, rank: &str) -> Html {
        html! {
            <tr>
                <td>{challenge}</td>
                <td>{rank}</td>
            </tr>
        }
    }

    fn rank_text(rank: i8) -> &'static str {
        match rank {
            4 => "S",
            3 => "A",
            2 => "B",
            1 => "C",
            _ => "Unknown",
        }
    }

    fn process(files: Option<FileList>) -> Msg {
        let Some(files) = files else {
            return Msg::Failure("No files selected".into());
        };

        let Ok(Some(iter)) = js_sys::try_iter(&files) else {
            return Msg::Failure("Unable to create iterator".into());
        };

        let Ok(values) = iter.collect::<Result<Vec<JsValue>, JsValue>>() else {
            return Msg::Failure("Unable to unpack values".into());
        };

        if values.len() != 1 {
            return Msg::Failure("Expected a single file".into());
        }

        values
            .into_iter()
            .map(web_sys::File::from)
            .map(gloo::file::File::from)
            .map(Msg::Upload)
            .next()
            .unwrap()
    }

    fn loaded(&mut self, name: String, data: String) -> bool {
        self.file = Some(FileDetails { data, name });
        self.reader = None;
        true
    }

    fn upload(&mut self, file: File, ctx: &Context<App>) -> bool {
        let link = ctx.link().clone();
        let name = file.name().clone();

        let task = read_as_text(&file, move |res| {
            let data = res.expect("failed to read file");
            link.send_message(Msg::Loaded(name, data))
        });

        self.reader = Some(task);
        true
    }
}

fn main() {
    yew::Renderer::<App>::new().render();
}
